__all__ = ['Command', 'Information', 'PostProcess', 'VTK', 'Foam', 'CaseBase', 'Conversion', 'Data', 'Envelope', 'SMTP', 'Figure', 'Timer', 'Version']


from .app.command.core import Command
from .app.information.core import Information
from .app.postprocess.core import PostProcess, VTK
from .base.core import Foam
from .util.object.case import CaseBase
from .util.object.conversion import Conversion
from .util.object.data import Data
from .util.object.email import Envelope, SMTP
from .util.object.figure import Figure
from .util.object.timer import Timer
from .util.object.version import Version



__all__ = ['Array', 'CachedLib', 'Dict', 'Document', 'FoamData', 'FoamItem', 'Keys', 'List', 'Location', 'Path', 'TupleSequence']


import pathlib as p
import types
import typing as t

if t.TYPE_CHECKING:
    import numpy as _numpy


CachedLib = t.Union[types.ModuleType, object, t.Callable]
Dict = t.Dict[str, t.Any]
List = t.List[t.Any]
Document = FoamItem = t.Union[Dict, List]
FoamData = t.List[FoamItem]
Location = t.Tuple[float, float, float]
Path = t.Union[str, p.Path]


class Array:
    # TODO: https://github.com/ramonhagenaars/nptyping?

    def __class_getitem__(cls, dimensions: 'Keys[int]') -> t.Union['_numpy.number', '_numpy.ndarray']:
        if not isinstance(dimensions, tuple):
            return cls.__class_getitem__((dimensions, ))
        assert all(isinstance(d, int) and d>=0 for d in dimensions)

        if dimensions == (0, ):
            return '_numpy.number'
        else:
            return '_numpy.ndarray'


class Keys:
    # TODO: https://stackoverflow.com/questions/47190218/proper-type-hint-for-getitem

    def __class_getitem__(cls, T: type) -> type:
        return t.Union[T, TupleSequence[T]]


class TupleSequence:
    # Reference: https://github.com/python/mypy/issues/184

    def __class_getitem__(cls, T: type) -> type:
        return t.Tuple[T, ...]



__all__ = ['copytree']


import shutil
import sys


def copytree(
    src, dst,
    symlinks=False, ignore=None, copy_function=shutil.copy2,
    ignore_dangling_symlinks=False, dirs_exist_ok=False,
):
    kwargs = {
        'symlinks': symlinks, 'ignore': ignore, 'copy_function': copy_function,
        'ignore_dangling_symlinks': ignore_dangling_symlinks,
    }
    if sys.version_info >= (3, 8):
        kwargs['dirs_exist_ok'] = dirs_exist_ok
    return shutil.copytree(src, dst, **kwargs)



__all__ = ['Parser']


import typing as t

from .lark import Lark
from .static import Static
from .url import Url
from .yaml import YAML
from ..base.type import Path

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ..base.core import Foam


class Parser:
    '''All parsers'''

    def __init__(self, static: Static, url: Url, yaml: YAML) -> None:
        self.static = static
        self.url = url
        self.yaml = yaml

    @classmethod
    def from_foam(cls, foam: 'Foam') -> 'Self':
        return cls(
            Static.from_foam(foam),
            Url.from_foam(foam),
            YAML.default(),
        )

    @classmethod
    def to_foam(cls, path: Path, **kwargs: t.Any) -> 'Foam':
        from ..base.core import Foam

        data = Lark.from_path(path, **kwargs).to_foam_data()
        data[0]['version'] = Foam.__version__.to_string()
        return Foam(data, path)



__all__ = ['CaseBase', 'CaseParameter']


import abc
import copy
import pathlib as p
import typing as t

from .data import Data
from ..function import dict_without_keys
from ...base.type import Dict, Path

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ...base.core import Foam


class CaseBase(abc.ABC):
    '''Case abstract class

    Example:
        ```
        class CaseCavity(CaseBase):
            __template__ = 'foam/static/demo/7/cavity.yaml'

            @property
            def t(self) -> float:
                return self._required['t']

            def optional(self):
                pass

            def required(self, t: float, dt: float):
                return self.dict_without_keys(vars(), 'self')

            def finalize(self, foam, optional, required):
                foam['foam'].set_via_dict({
                    'system': {
                        'controlDict': {
                            'endTime': self.t,
                            'deltaT': required['dt'],
                        },
                    },
                })
        ```

        >>> case_template = CaseCavity.new(t=0.3, dt=0.005)
        >>> print(case_template)
        CaseCavity(t=0.3, dt=0.005) \\
            .set_optional() \\
            .set_optional(t=0.3, dt=0.005)
        >>> for t in [0.3, 0.4, 0.5]:
        ...     parts = 'case', str(t)
        ...     case = case_template \\
        ...         .copy(deepcopy=False) \\
        ...         .set_required(t=t) \\
        ...         .save(*parts)
        ...     case.parameter.dump_to_path(*parts, 'meta.json')
        ...     codes = case.foam.cmd.all_run(overwrite=False)
        ...     print(codes)
        [0, 0]
        [0, 0]
        [0, 0]
    '''

    __template__ = None

    dict_without_keys = staticmethod(dict_without_keys)

    def __init__(self, **kwargs: t.Any) -> None:
        from ...base.core import Foam

        self._foam = Foam.from_path(self.tempalte)
        self._kwargs = kwargs
        self._optional = self.optional() or {}
        self._required = self.required(**kwargs) or {}

    def __repr__(self) -> str:
        kwargs = ', '.join(f'{k}={v!r}' for k, v in self._kwargs.items())
        optional = ', '.join(f'{k}={v!r}' for k, v in self._optional.items())
        required = ', '.join(f'{k}={v!r}' for k, v in self._required.items())
        return f'{self.__class__.__name__}({kwargs}) \\\n' \
            f'    .set_optional({optional}) \\\n' \
            f'    .set_optional({required})'

    @classmethod
    def new(cls, **kwargs: t.Any) -> 'Self':
        return cls(**kwargs)

    @abc.abstractmethod
    def optional(self) -> t.Optional[Dict]:
        '''Initialize optional properties'''
        pass

    @abc.abstractmethod
    def required(self, **kwargs: t.Any) -> t.Optional[Dict]:
        '''Initialize required properties'''
        pass

    @abc.abstractmethod
    def finalize(self, foam: 'Foam', optional: Dict, required: Dict) -> t.Optional['Foam']:
        '''Convert class properties to case parameters'''
        pass

    @property
    def foam(self) -> 'Foam':
        return self._foam

    @property
    def data(self) -> 'Data':
        return self._foam['foam']

    @property
    def tempalte(self) -> p.Path:
        assert self.__template__ is not None, f'Can\'t instantiate abstract class {self.__class__.__name__} with abstract variable __template__'

        return p.Path(self.__template__)

    @property
    def parameter(self) -> 'CaseParameter':
        return CaseParameter(self._optional, self._required)

    def copy(self, deepcopy: bool = False) -> 'Self':
        if deepcopy:
            return copy.deepcopy(self)
        else:
            return self.__class__(**self._kwargs)

    def set_from_path(self, *parts: str) -> 'Self':
        parameter = CaseParameter.load_from_path(*parts)
        self.set_optional(**parameter.optional)
        self.set_required(**parameter.required)
        return self

    def set_optional(self, **kwargs: t.Any) -> 'Self':
        self._optional.update(kwargs)
        return self

    def set_required(self, **kwargs: t.Any) -> 'Self':
        self._required.update(kwargs)
        return self

    def save(self, *parts: str) -> 'Self':
        foam = self.finalize(self._foam, self._optional, self._required)
        if foam is not None:
            self._foam = foam
        self._foam.save(p.Path(*map(str, parts)))
        return self


class CaseParameter(t.NamedTuple):
    '''Parameter for case'''

    optional: Dict
    required: Dict

    @classmethod
    def load_from_path(cls, *parts: str) -> 'Self':
        data = Data.load_from_path(*parts)
        return cls(optional=data.get('optional', {}), required=data.get('required', {}))

    def dump(self, *paths: Path) -> 'Self':
        self._data().dump(*paths)
        return self

    def dump_to_path(self, *parts: str) -> 'Self':
        self._data().dump_to_path(*parts)
        return self

    def _data(self) -> 'Data':
        return Data.from_dict({'optional': self.optional, 'required': self.required})



__all__ = ['Static']


import io
import pathlib as p
import shutil
import typing as t

from ..base.lib import py7zr
from ..base.type import Dict, Keys
from ..util.decorator import Match
from ..util.object.conversion import Conversion
from ..util.object.data import Data

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ..base.core import Foam


class Static:
    '''OpenFOAM static parser

    Example:
        >>> foam = Foam.from_demo('cavity')
        Foam.from_file('.../of.yaml/foam/static/demo/7/cavity.yaml', warn=False)
        >>> foam.save('case')

        >>> data = {
        ...     'name': 'Allrun',
        ...     'type': ['embed', 'text'],
        ...     'permission': 777,
        ...     'data': 'echo hello world!\n',
        ... }
        >>> static = Static.from_foam(foam)
        >>> static[tuple(data['type'])](data)
    '''

    match = Match.default()

    def __init__(self, foam: 'Foam') -> None:
        self._foam = foam

    def __getitem__(self, keys: Keys[str]) -> t.Callable:
        try:
            return lambda *args, **kwargs: self.match[keys](self, *args, **kwargs)
        except KeyError:
            raise Exception(f'Unknown types "{keys}"')

    @classmethod
    def from_foam(cls, foam: 'Foam') -> 'Self':
        return cls(foam)

    @match.register()
    def _(self, static: Dict) -> None:
        # do nothing
        pass

    @match.register('embed', 'text')
    def _(self, static: Dict) -> None:
        out = self._out(static['name'])
        self._foam._write(out, static['data'], static.get('permission', None))

    @match.register('embed', 'binary')
    def _(self, static: Dict) -> None:
        out = self._out(static['name'])
        out.write_bytes(static['data'])

    @match.register('embed', '7z')
    def _(self, static: Dict) -> None:
        out = self._out(static['name'])
        with py7zr.SevenZipFile(io.BytesIO(static['data']), mode='r') as z:
            z.extractall(path=out.parent)

    @match.register('path', 'raw')
    def _(self, static: Dict) -> None:
        out, in_ = self._out(static['name']), self._in(static['data'])
        if in_.is_dir():
            shutil.copytree(in_, out, dirs_exist_ok=True)
        elif in_.is_file():
            shutil.copyfile(in_, out)
        else:
            raise Exception('Target is neither a file nor a directory')

    @match.register('path', '7z')
    def _(self, static: Dict) -> None:
        out, in_ = self._out(static['name']), self._in(static['data'])
        with py7zr.SevenZipFile(in_, mode='r') as z:
            z.extractall(path=out.parent)

    @match.register('path', 'foam', 'json')
    def _(self, static: Dict) -> None:
        self._path_foam(static)

    @match.register('path', 'foam', 'toml')
    def _(self, static: Dict) -> None:
        self._path_foam(static)

    @match.register('path', 'foam', 'yaml')
    def _(self, static: Dict) -> None:
        self._path_foam(static)

    def _out(self, name: str) -> p.Path:
        out = self._foam._path(name)
        out.parent.mkdir(parents=True, exist_ok=True)
        return out

    def _in(self, data: str) -> p.Path:
        return self._foam._root / data

    def _path_foam(self, static: Dict) -> None:
        data = Data.from_dict()
        out, in_ = self._foam._path(), self._in(static['data'])
        data[static['name'].split('/')] = Conversion \
            .from_bytes(in_.read_bytes(), static['type'][2], all=False) \
            .to_document()
        self._foam.__class__([{'order': ['meta', 'foam']}, data._data], in_.parent, warn=False) \
            .save(out, paraview=False)



__all__ = ['classproperty', 'Match']


import functools as f
import typing as t

from ..base.lib import classproperty
from ..base.type import Keys, TupleSequence

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Match:
    '''Use decorator to simulate match syntax

    Example:
        >>> match = Match.default()

        >>> @match.register()
        ... def _():
        ...     return 0

        >>> @match.register('x')
        ... def _():
        ...     return 1

        >>> @match.register('x', 'y')
        ... def _():
        ...     return 2

        >>> match[()]()
        0
        >>> match['x']()
        1
        >>> match['x', 'y']()
        2
    '''

    def __init__(self) -> None:
        self._methods: t.Dict[TupleSequence[str], t.Callable] = {}

    def __getitem__(self, keys: Keys[str]) -> t.Callable:
        if not isinstance(keys, tuple):
            return self.__getitem__((keys, ))
        return self._methods[keys]

    @classmethod
    def default(cls) -> 'Self':
        return cls()

    def get(self, *keys: str, default: t.Optional[t.Callable] = None) -> t.Optional[t.Callable]:
        return self._methods.get(keys, default)

    def register(self, *types: str) -> t.Callable:
        '''Register methods for different types'''

        def decorate(func: t.Callable) -> t.Callable:
            self._methods[types] = func

            @f.wraps(func)
            def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
                return func(*args, **kwargs)

            return wrapper

        return decorate



import functools as _functools
import shutil as _shutil

from . import compat as _compat


_functools.cached_property = _compat.functools.cached_property
_functools.singledispatchmethod = _compat.functools.singledispatchmethod
_shutil.copytree = _compat.shutil.copytree


import importlib as _importlib

from .namespace import *


__import__ = lambda name: _importlib.import_module(name, __name__)
__all__ = __import__('.namespace').__all__
__doc__ = Foam.__doc__
__license__ = __import__('.util').function.license(full_text=False)
__version__ = Foam.__version__



__all__ = ['Foam']


import copy
import functools as f
import gc
import os
import pathlib as p
import typing as t
import urllib.parse
import urllib.request
import warnings as w

from .type import Dict, FoamData, Path
from ..parse import Parser
from ..util.object.conversion import Conversion
from ..util.object.data import Data
from ..util.object.version import Version

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ..app.command.core import Command
    from ..app.information.core import Information
    from ..app.postprocess.core import PostProcess


class Foam:
    '''Convert multiple dictionary type data to OpenFOAM test case

    Example:
        >>> foam = Foam.from_remote_demo('cavity')
        >>> foam['foam']['system', 'controlDict', 'endTime'] = 1.0
        >>> foam.save('cavity')
        >>> foam.cmd.all_run()
    '''

    __version__ = Version.from_string('0.12.5')

    def __init__(self, data: FoamData, root: Path, warn: bool = True) -> None:
        self._list = data
        self._root = p.Path(root)
        self._dest: t.Optional[p.Path] = None

        self._parser: t.Optional[Parser] = None
        self._cmd: t.Optional['Command'] = None
        self._info: t.Optional['Information'] = None
        self._post: t.Optional['PostProcess'] = None

        if warn:
            openfoam = set(map(str, self.meta.get('openfoam', [])))
            if str(self.environ['WM_PROJECT_VERSION']) not in openfoam:
                w.warn(f'OpenFOAM version mismatch: {root}')
            version = Version.from_string(str(self.meta.get('version', '0.0.0')))
            if self.__version__ < version:
                w.warn('Forward compatibility is not yet guaranteed')
            elif self.__version__ > version:
                w.warn('Backward compatibility is not yet guaranteed')

    def __getitem__(self, key: str) -> t.Optional[Data]:
        try:
            return Data.from_any(self._list[self.meta['order'].index(key)])
        except ValueError:
            return None

    def __repr__(self) -> str:
        return f'Foam({self._list!r}, {self._root!r})'

    def __str__(self) -> str:
        return f'<Foam @ "{self._root.absolute().as_posix()}">'

    @classmethod
    def list_demos(cls) -> t.List[str]:
        root = p.Path(__file__).parents[1] / 'static' / 'demo' / os.environ['WM_PROJECT_VERSION']
        return sorted(
            path.stem
            for path in root.iterdir()
            if path.suffix == '.yaml'
        )

    @classmethod
    def from_demo(cls, name: str = 'cavity', warn: bool = False) -> 'Self':
        version = os.environ['WM_PROJECT_VERSION']
        name = name if name.endswith('.yaml') else f'{name}.yaml'
        path = p.Path(__file__).parents[1] / 'static' / 'demo' / version / name
        try:
            self = cls.from_path(path, warn=warn)
        except FileNotFoundError:
            raise FileNotFoundError(f'No such demo: "{name[:-5]}" not in {cls.list_demos()}')
        else:
            print(f'Foam.from_path(\'{path.as_posix()}\', warn={warn})')
            self.meta.setdefault('openfoam', []).append(version)
            self.meta['version'] = cls.__version__.to_string()
            return self

    @classmethod
    def from_demos(cls, warn: bool = False) -> t.List['Self']:
        return list(map(lambda name: cls.from_demo(name, warn), cls.list_demos()))

    @classmethod
    def from_remote_demo(cls, name: str = 'cavity', timeout: t.Optional[float] = None, warn: bool = False) -> 'Self':
        version = os.environ['WM_PROJECT_VERSION']
        name = name if name.endswith('.yaml') else f'{name}.yaml'
        url = f'https://raw.githubusercontent.com/iydon/of.yaml/main/foam/static/demo/{version}/{name}'
        try:
            self = cls.from_remote_path(url, timeout, warn=warn)
        except Exception:  # urllib.error.URLError
            raise FileNotFoundError(f'No such demo: "{url}"')
        else:
            print(f'Foam.from_remote_path(\'{url}\', warn={warn})')
            self.meta.setdefault('openfoam', []).append(version)
            self.meta['version'] = cls.__version__.to_string()
            return self

    @classmethod
    def from_remote_demos(cls, timeout: t.Optional[float] = None, warn: bool = False) -> t.List['Self']:
        return list(map(lambda name: cls.from_remote_demo(name, timeout, warn), cls.list_demos()))

    @classmethod
    def from_remote_path(cls, url: str, timeout: t.Optional[float] = None, warn: bool = True) -> 'Self':
        with urllib.request.urlopen(url, timeout=timeout) as f:
            text = f.read()
        split_url = urllib.parse.urlsplit(url)
        path = p.Path(split_url.path)
        self = cls.from_text(text, '.', path.suffix[1:], warn=warn)
        self.parser.url.set_split_url(split_url)
        for old in self['static'] or []:
            types = tuple(old.get('type', []))
            old.update(self.parser.url[types](old.copy()))
        return self

    @classmethod
    def from_path(cls, path: Path, warn: bool = True) -> 'Self':
        '''Supported path mode: file, directory'''
        path = p.Path(path)
        if path.is_file():
            return cls.from_text(path.read_text(), path.parent, path.suffix[1:], warn=warn)
        elif path.is_dir():
            return cls.from_openfoam(path)
        else:
            raise Exception(f'Mode "{path.stat().st_mode}" is not supported')

    @classmethod
    def from_openfoam(cls, path: Path, **kwargs: t.Any) -> 'Self':
        '''From OpenFOAM directory'''
        return Parser.to_foam(path, **kwargs)

    @classmethod
    def from_text(
        cls,
        text: t.Union[bytes, str], root: Path,
        suffix: t.Optional[str] = None, warn: bool = True,
    ) -> 'Self':
        '''Supported formats: please refer to `Conversion`'''
        content = text if isinstance(text, bytes) else content.encode()
        if suffix is not None:
            data = Conversion.from_bytes(content, suffix, all=True).to_document()
        else:
            data = Conversion.auto_from_bytes(content, all=True).to_document()
        return cls(data, root, warn=warn)

    @classmethod
    def from_yaml(cls, text: str, root: Path, warn: bool = True) -> 'Self':
        data = Conversion.from_string(text, 'yaml', all=True).to_document()
        return cls(data, root, warn=warn)

    @classmethod
    def as_placeholder(cls) -> 'Self':
        return cls([{}], '', warn=False)

    @property
    def data(self) -> Data:
        return Data.from_list(self._list)

    @property
    def meta(self) -> Dict:
        '''Meta information'''
        return self._list[0]

    @property
    def parser(self) -> Parser:
        '''All parsers'''
        if self._parser is None:
            self._parser = Parser.from_foam(self)
        return self._parser

    @property
    def cmd(self) -> 'Command':
        '''`app.command.Command`'''
        from ..app.command.core import Command

        if self._cmd is None:
            self._cmd = Command.from_foam(self)
        return self._cmd

    @property
    def info(self) -> 'Information':
        '''`app.information.Information`'''
        from ..app.information.core import Information

        if self._info is None:
            self._info = Information.from_foam(self)
        return self._info

    @property
    def post(self) -> 'PostProcess':
        '''`app.postprocess.PostProcess`'''
        from ..app.postprocess.core import PostProcess

        if self._post is None:
            self._post = PostProcess.from_foam(self)
        return self._post

    @property
    def destination(self) -> p.Path:
        assert self._dest is not None, 'Please call `Foam::save` method first'

        return self._dest

    @destination.setter
    def destination(self, dest: Path) -> None:
        self._dest = p.Path(dest)

    @f.cached_property
    def application(self) -> str:
        '''Inspired by `getApplication`

        - Reference:
            - foamDictionary -disableFunctionEntries -entry application -value system/controlDict
        '''
        for key, value in self['foam']['system', 'controlDict'].items():
            if key.startswith('application'):
                return value
        raise Exception('Application not found')

    @f.cached_property
    def number_of_processors(self) -> int:
        '''Inspired by `getNumberOfProcessors`

        - Reference:
            - foamDictionary -disableFunctionEntries -entry numberOfSubdomains -value system/decomposeParDict
        '''
        try:
            return self['foam']['system', 'decomposeParDict', 'numberOfSubdomains']
        except Exception:
            return 1

    @f.cached_property
    def pipeline(self) -> t.List[t.Union[str, Dict]]:
        return (self['other'] or {}).get('pipeline', [])

    @f.cached_property
    def environ(self) -> t.Dict[str, str]:
        '''OpenFOAM environments'''
        return {
            key: value
            for key, value in os.environ.items()
            if any(key.startswith(p) for p in ['FOAM_', 'WM_'])
        }

    @f.cached_property
    def fields(self) -> t.Set[str]:
        return {v['FoamFile']['object'] for v in self['foam']['0'].values()}

    @f.cached_property
    def ndim(self) -> t.Optional[int]:
        # TODO: verify that this method is reliable
        system = self['foam']['system']
        block_mesh = system.get('blockMeshDict', None)
        if block_mesh is None:
            return None  # unknown ndim
        count = 3
        for block in block_mesh['blocks']:
            start = block.find(')')
            if start < 0:
                return None
            begin, end = block.find('(', start+1), block.find(')', start+1)
            if begin < 0 or end < 0:
                return None
            grids = block[begin+1: end].split()
            count = min(count, len(grids)-grids.count('1'))
        return count

    def copy(self) -> 'Self':
        other = self.__class__(copy.deepcopy(self._list), self._root)
        other._dest = self._dest
        return other

    def save(self, dest: Path, paraview: bool = True) -> 'Self':
        '''Persist case to hard disk'''
        self._dest = p.Path(dest)
        self._dest.mkdir(parents=True, exist_ok=True)
        self._save_foam()
        self._save_static()
        if paraview:
            self._write(self._dest/'paraview.foam', '')
        return self

    def reset(self) -> 'Self':
        self._dest = self._cmd = self._info = self._post = None
        for obj in gc.get_objects():
            if isinstance(obj, f._lru_cache_wrapper):
                obj.cache_clear()
        return self

    def _write(self, path: p.Path, string: str, permission: t.Optional[int] = None) -> None:
        with open(path, 'w', encoding='utf-8', newline='\n') as f:  # CRLF -> LF
            f.write(string)
        if permission is not None:
            path.chmod(int(str(permission), base=8))

    def _save_static(self) -> None:
        # TODO: add to parse sub-module
        for static in self['static'] or []:
            self.parser.static[tuple(static['type'])](static)

    def _save_foam(self) -> None:
        foam = self['foam']
        for keys, data in self._extract_files({} if foam is None else foam.data):
            # pre-process FoamFile to avoid duplicate descriptions (not recommended yet)
            if data['FoamFile'] is None:
                data.pop('FoamFile')
            elif isinstance(data['FoamFile'], str):
                data['FoamFile'] = {'class': data['FoamFile']}
            if 'FoamFile' in data:
                for key, value in [('version', 2.0), ('format', 'ascii'), ('object', keys[-1])]:
                    data['FoamFile'].setdefault(key, value)
            # write the parsed text data
            path = self._path(*map(str, keys))  # self._dest is not None
            path.parent.mkdir(parents=True, exist_ok=True)
            self._write(path, '\n'.join(self.parser.yaml.data(data)))

    def _extract_files(
        self,
        data: Dict, keys: t.List[str] = [],
    ) -> t.Iterator[t.Tuple[t.List[str], Dict]]:
        if 'FoamFile' in data:
            yield keys, data.copy()
        else:
            for key, value in data.items():
                yield from self._extract_files(value, keys+[key])

    def _path(self, *parts: str) -> p.Path:
        # TODO: use "prefix" will cause some of the `Command` methods to fail
        prefix = '.'  # (self['other'] or {}).get('directory', '.')
        return self.destination / prefix / p.Path(*parts)



__all__ = ['Conversion']


import json
import pathlib as p
import pickle
import typing as t

from ...base.lib import tomlkit, yaml
from ...base.type import Document, Path

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Conversion:
    '''Conversion between object and bytes/string

    Example:
        >>> data = {'a': 1, 'b': [2, 3], 'c': {'4': 5}}
        >>> print(Conversion.from_document(data).to_yaml())
        a: 1
        b:
        - 2
        - 3
        c:
        '4': 5
    '''

    types = {'json', 'pickle', 'toml', 'yaml'}

    def __init__(self, document: Document) -> None:
        self._document = document

    @classmethod
    def auto_from_bytes(cls, content: bytes, all: bool = False) -> 'Self':
        for type in cls.types:
            try:
                return cls.from_bytes(content, type, all)
            except Exception:
                pass
        raise Exception('Unable to recognize content type')

    @classmethod
    def auto_from_string(cls, text: str, all: bool = False) -> 'Self':
        return cls.auto_from_bytes(text.encode(), all)

    @classmethod
    def auto_from_path(cls, path: Path, all: bool = False) -> 'Self':
        return cls.auto_from_bytes(p.Path(path).read_bytes(), all)

    @classmethod
    def from_document(cls, document: Document) -> 'Self':
        return cls(document)

    @classmethod
    def from_bytes(cls, content: bytes, type: str = 'json', all: bool = False) -> 'Self':
        type = type.lstrip('.')
        for types, loads in [
            ({'json'}, lambda content: cls.from_json(content.decode())),
            ({'pickle', 'pkl'}, lambda content: cls.from_pickle(content)),
            ({'toml'}, lambda content: cls.from_toml(content.decode())),
            ({'yaml', 'yml'}, lambda content: cls.from_yaml(content.decode(), all)),
        ]:
            if type in types:
                return loads(content)
        raise Exception(f'"{type}" is not a valid type string')

    @classmethod
    def from_string(cls, text: str, type: str = 'json', all: bool = False) -> 'Self':
        return cls.from_bytes(text.encode(), type, all)

    @classmethod
    def from_path(cls, path: Path, all: bool = False) -> 'Self':
        path = p.Path(path)
        return cls.from_bytes(path.read_bytes(), path.suffix[1:], all)

    @classmethod
    def from_json(cls, text: str) -> 'Self':
        return cls(json.loads(text))

    @classmethod
    def from_pickle(cls, text: bytes) -> 'Self':
        return cls(pickle.loads(text))

    @classmethod
    def from_toml(cls, text: str) -> 'Self':
        return cls(tomlkit.loads(text))

    @classmethod
    def from_yaml(cls, text: str, all: bool = False) -> 'Self':
        document = list(yaml.load_all(text)) if all else yaml.load(text)
        return cls(document)

    def to_document(self) -> Document:
        return self._document

    def to_bytes(self, type: str = 'json', all: bool = False, **kwargs: t.Any) -> bytes:
        type = type.lstrip('.')
        for types, dumps in [
            ({'json'}, lambda: self.to_json(**kwargs).encode()),
            ({'pickle', 'pkl'}, lambda: self.to_pickle(**kwargs)),
            ({'toml'}, lambda: self.to_toml(**kwargs).encode()),
            ({'yaml', 'yml'}, lambda: self.to_yaml(all, **kwargs).encode()),
        ]:
            if type in types:
                return dumps()
        raise Exception(f'"{type}" is not a valid type string')

    def to_string(self, type: str = 'json', all: bool = False, **kwargs: t.Any) -> str:
        return self.to_bytes(type, all, **kwargs).decode()

    def to_path(self, path: Path, all: bool = False, **kwargs: t.Any) -> p.Path:
        path = p.Path(path)
        path.write_bytes(self.to_bytes(path.suffix[1:], all, **kwargs))
        return path

    def to_json(self, **kwargs: t.Any) -> str:
        kwargs = {'ensure_ascii': False, **kwargs}
        return json.dumps(self._document, **kwargs)

    def to_pickle(self, **kwargs: t.Any) -> bytes:
        return pickle.dumps(self._document, **kwargs)

    def to_toml(self, **kwargs: t.Any) -> str:
        return tomlkit.dumps(self._document, **kwargs)

    def to_yaml(self, all: bool = False, **kwargs: t.Any) -> str:
        kwargs = {'indent': 4, **kwargs}
        return (yaml.dump_all if all else yaml.dump)(self._document, **kwargs)



__all__ = ['Information']


import collections as c
import os
import pathlib as p
import re
import typing as t

from ...base.type import Path

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ..command.core import Command
    from ...base.core import Foam


class Information:
    '''OpenFOAM information wrapper'''

    def __init__(self, foam: 'Foam') -> None:
        self._foam = foam
        self._cmd: t.Optional['Command'] = None

    @classmethod
    def from_foam(cls, foam: 'Foam') -> 'Self':
        return cls(foam)

    @classmethod
    def from_nothing(cls) -> 'Self':
        from ...base.core import Foam

        return cls(Foam.as_placeholder())

    @property
    def cmd(self) -> 'Command':
        '''Command without asserting (no need to call `Foam::save` method first)'''
        from ..command.core import Command

        if self._cmd is None:
            self._cmd = Command.from_foam_without_asserting(self._foam)
        return self._cmd

    @property
    def environ(self) -> t.Dict[str, str]:
        '''OpenFOAM environments (aliase for `Foam::environ` property)'''
        if self._foam is not None:  # True
            return self._foam.environ
        else:
            return {
                key: value
                for key, value in os.environ.items()
                if any(key.startswith(p) for p in ['FOAM_', 'WM_'])
            }

    @property
    def root(self) -> p.Path:
        return p.Path(self.environ['WM_PROJECT_DIR'])

    @property
    def shared_libraries(self) -> t.Set[str]:
        ans = set()
        pattern = re.compile(fr'{self.environ["FOAM_LIBBIN"]}.+?\.so')
        for command in self.commands(foam_only=True):
            path = self.cmd.which(command)
            if path is not None:
                stdout = self.cmd.raw(f'ldd {path}').stdout.decode()
                ans.update(pattern.findall(stdout))
        return ans

    def search(self, *targets: str, process: bool = True) -> t.Union[str, t.Set[str]]:
        '''`foamSearch` wrapper

        - Reference:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/bin/foamSearch
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/miscellaneous/foamDictionary/foamDictionary.C
        '''
        assert len(targets) > 1

        fields = '.'.join(map(lambda string: string.replace(' ', ''), targets[1:]))
        command = f'foamSearch {self.environ["FOAM_TUTORIALS"]} {targets[0]} "{fields}"'
        stdout = self.cmd.raw(command, output=True).stdout.decode()
        if not process:
            return stdout
        try:
            return set(
                line.split(' ', maxsplit=1)[-1].strip()
                for line in stdout.splitlines()
            )
        except Exception:
            return stdout

    def search_yaml(self, *targets: str, root: Path = '.') -> t.Dict[t.Hashable, t.Set[str]]:
        '''`foamSearch` in YAML

        - Note:
            - `targets` should be as detailed as possible, as it is assumed that `targets` will only appear once in a file
        '''
        from ...base.core import Foam

        assert targets

        record = c.defaultdict(set)
        hashing = lambda string: string.lower().replace(' ', '')
        hashed_targets = tuple(map(hashing, targets))
        length = len(targets)
        for path in p.Path(root).rglob('*'):
            if path.suffix in {'.yaml', '.yml'}:
                # try-except or try-except-else?
                try:
                    foam = Foam.from_file(path, warn=False)
                    for keys, _ in foam['foam'].items(with_list=False):
                        hashed_keys = tuple(map(hashing, keys))
                        if hashed_keys[-length:] == hashed_targets:
                            record[foam['foam'][keys]].add(path.absolute().as_posix())
                            break
                except Exception:  # TODO: catch specific exceptions only
                    pass
        return dict(record)

    def commands(self, foam_only: bool = True) -> t.Set[str]:
        strize = lambda command: {
            str: lambda x: x,
            dict: lambda x: x['command'],
        }[type(command)](command)
        func = lambda x: self.cmd._split(strize(x), False)[0]
        commands = set(map(func, self._foam.pipeline))
        if foam_only:
            commands = {
                command for command in commands
                if self.environ['FOAM_APPBIN'] in (self.cmd.which(command) or '')
            }
        return commands



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['foam',
 'foam.app',
 'foam.app.command',
 'foam.app.information',
 'foam.app.postprocess',
 'foam.base',
 'foam.compat',
 'foam.parse',
 'foam.util',
 'foam.util.object']

package_data = \
{'': ['*'],
 'foam': ['static/*',
          'static/demo/7/*',
          'static/demo/8/*',
          'static/demo/9/*',
          'static/grammar/*']}

install_requires = \
['PyYAML>=6.0,<7.0']

extras_require = \
{'7z': ['py7zr>=0.17.2,<0.18.0'],
 'cli': ['click>=8.0.3,<9.0.0'],
 'full': ['click>=8.0.3,<9.0.0',
          'lark>=1.1.2,<2.0.0',
          'matplotlib>=3.5.3,<4.0.0',
          'py7zr>=0.17.2,<0.18.0',
          'tomlkit>=0.11.5,<0.12.0',
          'tqdm>=4.63.1,<5.0.0',
          'typing-extensions>=4.3.0,<5.0.0',
          'vtk>=9.1.0,<10.0.0'],
 'lark': ['lark>=1.1.2,<2.0.0'],
 'mpl': ['matplotlib>=3.5.3,<4.0.0'],
 'toml': ['tomlkit>=0.11.5,<0.12.0'],
 'tqdm': ['tqdm>=4.63.1,<5.0.0'],
 'type': ['typing-extensions>=4.3.0,<5.0.0'],
 'vtk': ['vtk>=9.1.0,<10.0.0']}

setup_kwargs = {
    'name': 'ifoam',
    'version': '0.12.5',
    'description': 'Python Interface to OpenFOAM Case (Configured Using YAML)',
    'long_description': '<!-- Template from https://github.com/othneildrew/Best-README-Template -->\n<div id="top"></div>\n\n\n\n<!-- PROJECT SHIELDS -->\n[![Contributors][contributors-shield]][contributors-url]\n[![Forks][forks-shield]][forks-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url]\n[![GPL-3.0 License][license-shield]][license-url]\n\n\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n  <a href="https://github.com/iydon/of.yaml">\n    🟢⬜🟩⬜🟩<br />\n    ⬜⬜⬜⬜⬜<br />\n    🟩⬜🟩⬜🟩<br />\n    ⬜⬜⬜⬜⬜<br />\n    🟩⬜🟩⬜🟩<br />\n  </a>\n\n  <h3 align="center">OpenFOAM.YAML</h3>\n\n  <p align="center">\n    Python Interface to OpenFOAM Case (Configured Using YAML)\n    <br />\n    <a href="https://ifoam.readthedocs.io"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    View <a href="https://github.com/iydon/of.yaml-template">Demo</a>/<a href="https://github.com/iydon/of.yaml-tutorial">Tutorial</a>\n    ·\n    <a href="https://github.com/iydon/of.yaml/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/iydon/of.yaml/issues">Request Feature</a>\n  </p>\n</div>\n\n\n\n<!-- TABLE OF CONTENTS -->\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#about-the-project">About The Project</a>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#installation">Installation</a></li>\n        <li><a href="#demo">Demo</a></li>\n      </ul>\n    </li>\n    <li><a href="#contributing">Contributing</a></li>\n    <li><a href="#license">License</a></li>\n    <li><a href="#contact">Contact</a></li>\n  </ol>\n</details>\n\n\n\n<!-- ABOUT THE PROJECT -->\n## About The Project\n\nThis repository was originally designed to solve the problem of complex OpenFOAM case structure, and the solution was to re-present the original cases using the common configuration file format YAML. Later, since there is a corresponding package for the YAML format in Python, I wrote this Python interface package for OpenFOAM, and then I added progress bars to most OpenFOAM solvers by analyzing log files in real time. Although there are still many details to be specified in this repository, its function of generating cases and calling solvers is ready for preliminary use, for example, I used this package to generate cases in batch in my own project. In the future I would like to integrate the post-processing steps into this interface package as well.\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- GETTING STARTED -->\n## Getting Started\n\nThis project currently uses Poetry to manage Python dependencies. I\'ve heard good things about [PDM](https://github.com/pdm-project/pdm) so far, and may provide PDM support subsequently.\n\n### Installation\n\n```sh\npip3 install ifoam[full]\n```\n\n### Demo\n\nSave the following demo code as a separate file (e.g. `demo.py`).\n\n```python\nfrom foam import Foam\n\nfoam = Foam.from_demo(\'cavity\')\nfoam[\'foam\'][\'system\', \'controlDict\', \'endTime\'] = 1.0\nfoam.save(\'cavity\')\nfoam.cmd.all_run()\n```\n\nRunning the demo code in the virtual environment results in the following output.\n\n```sh\n$ python demo.py\n\nFoam.from_file(\'.../of.yaml/foam/demo/7/cavity.yaml\', warn=False)\nRunning blockMesh on .../of.yaml/cavity using 1 processes if in parallel\nRunning icoFoam on .../of.yaml/cavity using 1 processes if in parallel\n100%|█████████████████████████████████████| 1.0/1.0 [00:02<00:00,  2.24s/it]\n```\n\n\n\n<!-- CONTRIBUTING -->\n## Contributing\n\nContributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repository and create a pull request. You can also simply open an issue with the tag "enhancement".\nDon\'t forget to give the project a star! Thanks again!\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the GPL-3.0 License. See `LICENSE.txt` for more information.\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- CONTACT -->\n## Contact\n\nIydon Liang - [@iydon](https://github.com/iydon) - liangiydon_AT_gmail.com\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- MARKDOWN LINKS & IMAGES -->\n[contributors-shield]: https://img.shields.io/github/contributors/iydon/of.yaml.svg?style=for-the-badge\n[contributors-url]: https://github.com/iydon/of.yaml/graphs/contributors\n[forks-shield]: https://img.shields.io/github/forks/iydon/of.yaml.svg?style=for-the-badge\n[forks-url]: https://github.com/iydon/of.yaml/network/members\n[stars-shield]: https://img.shields.io/github/stars/iydon/of.yaml.svg?style=for-the-badge\n[stars-url]: https://github.com/iydon/of.yaml/stargazers\n[issues-shield]: https://img.shields.io/github/issues-closed/iydon/of.yaml.svg?style=for-the-badge\n[issues-url]: https://github.com/iydon/of.yaml/issues\n[license-shield]: https://img.shields.io/github/license/iydon/of.yaml.svg?style=for-the-badge\n[license-url]: https://github.com/iydon/of.yaml/blob/master/LICENSE.txt\n',
    'author': 'Iydon Liang',
    'author_email': 'liangiydon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iydon/of.yaml',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



__all__ = ['dict_without_keys', 'grammar', 'license']


import pathlib as p


root = p.Path(__file__).parents[1]


def dict_without_keys(data: dict, *keys: str) -> dict:
    # dict := t.Dict[t.Hashable, t.Any]
    return {key: data[key] for key in data.keys() ^ keys}


def grammar() -> str:
    return (root/'static'/'grammar'/'openfoam.lark').read_text()


def license(full_text: bool = False) -> str:
    if full_text:
        return (root/'static'/'LICENSE.txt').read_text()
    else:
        return 'GPL-3.0-only'



__all__ = ['Default', 'Apps']


import re
import typing as t

from ...base.lib import tqdm

if t.TYPE_CHECKING:
    from ...base.core import Foam


class Default:
    '''Default app'''

    def __init__(self, foam: 'Foam') -> None:
        pass

    def __enter__(self) -> 'Default':
        return self

    def __exit__(self, type, value, traceback) -> None:
        pass

    def step(self, line: bytes) -> None:
        pass

    def close(self) -> None:
        pass


class AppBase(Default):
    '''Base app'''

    def __init__(self, foam: 'Foam') -> None:
        start = float(foam['foam']['system', 'controlDict', 'startTime'])
        end = float(foam['foam']['system', 'controlDict', 'endTime'])
        self.pbar = tqdm.tqdm(total=end-start)
        self._new = self._old = start

    def __exit__(self, type, value, traceback) -> None:
        self.pbar.close()

    def step(self, line: bytes) -> None:
        now = self.now(line) or self._new
        self._new, self._old = now, self._new
        self.pbar.update(self._new-self._old)

    def now(self, line: bytes) -> t.Optional[float]:
        raise NotImplementedError


class AppByTimeI(AppBase):
    '''
    - solver:
        - multiphase:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/cavitatingFoam/cavitatingFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/driftFluxFoam/driftFluxFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/compressibleMultiphaseInterFoam/compressibleMultiphaseInterFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/interPhaseChangeFoam/interPhaseChangeFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/potentialFreeSurfaceFoam/potentialFreeSurfaceFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/reactingEulerFoam/reactingMultiphaseEulerFoam/reactingMultiphaseEulerFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/reactingEulerFoam/reactingTwoPhaseEulerFoam/reactingTwoPhaseEulerFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/twoLiquidMixingFoam/twoLiquidMixingFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/compressibleInterFoam/compressibleInterFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/compressibleInterFoam/compressibleInterFilmFoam/compressibleInterFilmFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/interFoam/interFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/interFoam/interMixingFoam/interMixingFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/multiphaseEulerFoam/multiphaseEulerFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/twoPhaseEulerFoam/twoPhaseEulerFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/multiphase/multiphaseInterFoam/multiphaseInterFoam.C
        - discreteMethods:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/discreteMethods/molecularDynamics/mdFoam/mdFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/discreteMethods/molecularDynamics/mdEquilibrationFoam/mdEquilibrationFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/discreteMethods/dsmc/dsmcFoam/dsmcFoam.C
        - heatTransfer:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/heatTransfer/thermoFoam/thermoFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/heatTransfer/chtMultiRegionFoam/chtMultiRegionFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/heatTransfer/buoyantSimpleFoam/buoyantSimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/heatTransfer/buoyantPimpleFoam/buoyantPimpleFoam.C
        - combustion:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/reactingFoam/reactingFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/reactingFoam/rhoReactingFoam/rhoReactingFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/reactingFoam/rhoReactingBuoyantFoam/rhoReactingBuoyantFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/fireFoam/fireFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/XiFoam/XiFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/PDRFoam/PDRFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/chemFoam/chemFoam.C
        - DNS:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/DNS/dnsFoam/dnsFoam.C
        - incompressible:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/nonNewtonianIcoFoam/nonNewtonianIcoFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/adjointShapeOptimizationFoam/adjointShapeOptimizationFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/shallowWaterFoam/shallowWaterFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/boundaryFoam/boundaryFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/pimpleFoam/pimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/pimpleFoam/SRFPimpleFoam/SRFPimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/pisoFoam/pisoFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/simpleFoam/simpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/simpleFoam/SRFSimpleFoam/SRFSimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/simpleFoam/porousSimpleFoam/porousSimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/incompressible/icoFoam/icoFoam.C
        - electromagnetics:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/electromagnetics/mhdFoam/mhdFoam.C
        - lagrangian:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/reactingParcelFoam/reactingParcelFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/reactingParcelFoam/simpleReactingParcelFoam/simpleReactingParcelFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/coalChemistryFoam/coalChemistryFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/uncoupledKinematicParcelFoam/uncoupledKinematicParcelFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/DPMFoam/DPMFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/icoUncoupledKinematicParcelFoam/icoUncoupledKinematicParcelFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/sprayFoam/sprayFoam.C
        - compressible:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/compressible/rhoPimpleFoam/rhoPimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/compressible/rhoCentralFoam/rhoCentralFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/compressible/rhoSimpleFoam/rhoSimpleFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/compressible/rhoSimpleFoam/rhoPorousSimpleFoam/rhoPorousSimpleFoam.C
        - basic:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/basic/laplacianFoam/laplacianFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/basic/scalarTransportFoam/scalarTransportFoam.C
    - utility:
        - postProcessing:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/postProcess/postProcess.C
        - parallelProcessing:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/parallelProcessing/reconstructParMesh/reconstructParMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/parallelProcessing/reconstructPar/reconstructPar.C
    '''

    def now(self, line: bytes) -> t.Optional[float]:
        line = line.strip()
        if line.startswith(b'Time = '):
            return float(line[7:])


class AppByTimeII(AppBase):
    '''
    - utility:
        - postProcessing:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamToVTK/foamToVTK.C
    '''

    def now(self, line: bytes) -> t.Optional[float]:
        line = line.strip()
        if line.startswith(b'Time: '):
            return float(line[6:])


class AppByIterationI(AppBase):
    '''
    - solver:
        - electromagnetics:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/electromagnetics/electrostaticFoam/electrostaticFoam.C
    '''

    def now(self, line: bytes) -> t.Optional[float]:
        line = line.strip()
        if line.startswith(b'Iteration = '):
            return float(line[12:])


class AppByIterationII(AppBase):
    '''
    - solver:
        - stressAnalysis:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/stressAnalysis/solidEquilibriumDisplacementFoam/solidEquilibriumDisplacementFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/stressAnalysis/solidDisplacementFoam/solidDisplacementFoam.C
    '''

    def now(self, line: bytes) -> t.Optional[float]:
        line = line.strip()
        if line.startswith(b'Iteration: '):
            return float(line[11:])


class AppByProcessor(AppBase):
    '''
    - utility:
        - parallelProcessing:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/parallelProcessing/decomposePar/decomposePar.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/parallelProcessing/redistributePar/redistributePar.C
    '''

    def __init__(self, foam: 'Foam') -> None:
        start = 0
        end = foam.number_of_processors - 1
        self.pbar = tqdm.tqdm(total=end-start)
        self._new = self._old = start

    def now(self, line: bytes) -> t.Optional[float]:
        line = line.strip()
        if line.startswith(b'Processor ') and line[10:].isdigit():
            return int(line[10:])


class AppByOther(AppBase):
    '''
    - solver:
        - combustion:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/XiFoam/XiEngineFoam/XiEngineFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/combustion/coldEngineFoam/coldEngineFoam.C
        - electromagnetics:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/electromagnetics/magneticFoam/magneticFoam.C
        - lagrangian:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/DPMFoam/MPPICFoam/MPPICFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/lagrangian/sprayFoam/engineFoam/engineFoam.C
        - financial:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/financial/financialFoam/financialFoam.C
        - basic:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/solvers/basic/potentialFoam/potentialFoam.C
    - utility:
        - postProcessing:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/lagrangian/steadyParticleTracks/steadyParticleTracks.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/lagrangian/particleTracks/particleTracks.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/miscellaneous/engineCompRatio/engineCompRatio.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/miscellaneous/temporalInterpolate/temporalInterpolate.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/miscellaneous/pdfPlot/pdfPlot.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/miscellaneous/postChannel/postChannel.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/graphics/PVReaders/vtkPVFoam/vtkPVFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/graphics/PVReaders/vtkPVblockMesh/vtkPVblockMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/noise/noise.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamToTecplot360/foamToTecplot360.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/smapToFoam/smapToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamToGMV/foamToGMV.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamToTetDualMesh/foamToTetDualMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamDataToFluent/foamDataToFluent.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamToEnsight/foamToEnsight.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/postProcessing/dataConversion/foamToEnsightParts/foamToEnsightParts.C
        - mesh:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/removeFaces/removeFaces.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/modifyMesh/modifyMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/collapseEdges/collapseEdges.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/splitCells/splitCells.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/PDRMesh/PDRMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/refineWallLayer/refineWallLayer.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/refinementLevel/refinementLevel.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/autoRefineMesh/autoRefineMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/selectCells/selectCells.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/combinePatchFaces/combinePatchFaces.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/advanced/refineHexMesh/refineHexMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/foamToStarMesh/foamToStarMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/fluent3DMeshToFoam/fluent3DMeshToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/vtkUnstructuredToFoam/vtkUnstructuredToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/Optional/ccm26ToFoam/ccm26ToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/foamToSurface/foamToSurface.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/ideasUnvToFoam/ideasUnvToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/netgenNeutralToFoam/netgenNeutralToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/gmshToFoam/gmshToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/writeMeshObj/writeMeshObj.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/mshToFoam/mshToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/ansysToFoam/ansysToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/foamMeshToFluent/foamMeshToFluent.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/gambitToFoam/gambitToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/plot3dToFoam/plot3dToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/fluentMeshToFoam/fluentMeshToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/star4ToFoam/star4ToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/datToFoam/datToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/tetgenToFoam/tetgenToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/cfx4ToFoam/cfx4ToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/sammToFoam/sammToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/star3ToFoam/star3ToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/conversion/kivaToFoam/kivaToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/splitMeshRegions/splitMeshRegions.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/stitchMesh/stitchMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/createPatch/createPatch.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/polyDualMesh/polyDualMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/moveEngineMesh/moveEngineMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/moveMesh/moveMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/moveDynamicMesh/moveDynamicMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/renumberMesh/renumberMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/rotateMesh/rotateMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/deformedGeom/deformedGeom.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/objToVTK/objToVTK.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/zipUpMesh/zipUpMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/transformPoints/transformPoints.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/mergeOrSplitBaffles/mergeOrSplitBaffles.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/attachMesh/attachMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/setsToZones/setsToZones.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/singleCellMesh/singleCellMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/orientFaceZone/orientFaceZone.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/topoSet/topoSet.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/mergeMeshes/mergeMeshes.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/setSet/setSet.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/createBaffles/createBaffles.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/checkMesh/checkMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/autoPatch/autoPatch.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/refineMesh/refineMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/insideCells/insideCells.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/mirrorMesh/mirrorMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/splitMesh/splitMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/flattenMesh/flattenMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/manipulation/subsetMesh/subsetMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/extrude/extrudeMesh/extrudeMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/extrude/extrudeToRegionMesh/extrudeToRegionMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/snappyHexMesh/snappyHexMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/extrude2DMesh/extrude2DMesh/extrude2DMesh/extrude2DMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/blockMesh/blockMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/foamyMesh/foamyHexMeshSurfaceSimplify/foamyHexMeshSurfaceSimplify.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/foamyMesh/foamyHexMesh/foamyHexMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/foamyMesh/foamyQuadMesh/foamyQuadMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/foamyMesh/foamyHexMeshBackgroundMesh/foamyHexMeshBackgroundMesh.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/mesh/generation/foamyMesh/cellSizeAndAlignmentGrid/cellSizeAndAlignmentGrid.C
        - miscellaneous:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/miscellaneous/foamDictionary/foamDictionary.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/miscellaneous/foamListTimes/foamListTimes.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/miscellaneous/foamFormatConvert/foamFormatConvert.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/miscellaneous/patchSummary/patchSummary.C
        - thermophysical:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/thermophysical/chemkinToFoam/chemkinToFoam.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/thermophysical/equilibriumFlameT/equilibriumFlameT.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/thermophysical/mixtureAdiabaticFlameT/mixtureAdiabaticFlameT.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/thermophysical/equilibriumCO/equilibriumCO.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/thermophysical/adiabaticFlameT/adiabaticFlameT.C
        - preProcessing:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/createExternalCoupledPatchGeometry/createExternalCoupledPatchGeometry.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/changeDictionary/changeDictionary.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/applyBoundaryLayer/applyBoundaryLayer.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/mdInitialise/mdInitialise.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/faceAgglomerate/faceAgglomerate.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/wallFunctionTable/wallFunctionTable.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/engineSwirl/engineSwirl.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/mapFieldsPar/mapFieldsPar.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/setWaves/setWaves.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/dsmcInitialise/dsmcInitialise.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/foamSetupCHT/foamSetupCHT.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/setFields/setFields.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/viewFactorsGen/viewFactorsGen.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/mapFields/mapFields.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/boxTurb/boxTurb.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/preProcessing/foamUpgradeCyclics/foamUpgradeCyclics.C
        - surface:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceConvert/surfaceConvert.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceAutoPatch/surfaceAutoPatch.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceTransformPoints/surfaceTransformPoints.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceMeshExport/surfaceMeshExport.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceOrient/surfaceOrient.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceMeshConvertTesting/surfaceMeshConvertTesting.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceMeshConvert/surfaceMeshConvert.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceSplitByTopology/surfaceSplitByTopology.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceSplitNonManifolds/surfaceSplitNonManifolds.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceFeatureExtract/surfaceFeatureExtract.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceFeatures/surfaceFeatures.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceToPatch/surfaceToPatch.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceCheck/surfaceCheck.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceRedistributePar/surfaceRedistributePar.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceRefineRedGreen/surfaceRefineRedGreen.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceLambdaMuSmooth/surfaceLambdaMuSmooth.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceSplitByPatch/surfaceSplitByPatch.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceMeshInfo/surfaceMeshInfo.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceAdd/surfaceAdd.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceMeshImport/surfaceMeshImport.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceHookUp/surfaceHookUp.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceSubset/surfaceSubset.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceFeatureConvert/surfaceFeatureConvert.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceClean/surfaceClean.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceCoarsen/surfaceCoarsen.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfacePointMerge/surfacePointMerge.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceBooleanFeatures/surfaceBooleanFeatures.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceInertia/surfaceInertia.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceFind/surfaceFind.C
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/applications/utilities/surface/surfaceMeshTriangulate/surfaceMeshTriangulate.C
    '''

    def now(self, line: bytes) -> t.Optional[float]:
        raise NotImplementedError


Apps = {}
if tqdm.is_available():
    pattern = re.compile(r'\w+(?=\.C)')
    for App in [AppByTimeI, AppByTimeII, AppByIterationI, AppByIterationII, AppByProcessor]:
        for name in pattern.findall(App.__doc__):
            Apps[name] = App



__all__ = ['Data']


import pathlib as p
import typing as t

from .conversion import Conversion
from ...base.type import Dict, FoamItem, Keys, List, Path

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Data:
    '''Multi-key dictionary

    Example:
        >>> data = Data.from_dict_keys(
        ...     ('left', 'x'), ('left', 'y'),
        ...     ('right', 'x'), ('right', 'y'),
        ...     default=list,
        ... )
        >>> data['left', 'x'].append({...})
        >>> data['right', 'y'].append({...})

        >>> for key, val in data.items():
        ...     print(key, val)
        ('left', 'x') [{Ellipsis}]
        ('left', 'y') []
        ('right', 'x') []
        ('right', 'y') [{Ellipsis}]
    '''

    def __init__(self, data: FoamItem) -> None:
        self._data = data

    def __contains__(self, keys: Keys[t.Any]) -> bool:
        if isinstance(keys, tuple):
            ans = self._data
            for key in keys:
                if key not in ans:
                    return False
                else:
                    ans = ans[key]
            return True
        elif isinstance(keys, list):
            return self.__contains__(tuple(keys))
        else:
            return self._data.__contains__(keys)

    def __getitem__(self, keys: Keys[t.Any]) -> t.Any:
        if isinstance(keys, tuple):
            ans = self._data
            for key in keys:
                ans = ans[key]
            return ans
        elif isinstance(keys, list):
            # TODO: DeprecationWarning
            return self.__getitem__(tuple(keys))
        else:
            return self._data[keys]

    def __setitem__(self, keys: Keys[t.Any], value: t.Any) -> None:
        if isinstance(keys, tuple):
            assert keys

            ans = self._data
            for key in keys[:-1]:
                if isinstance(ans, dict):
                    ans = ans.setdefault(key, {})
                elif isinstance(ans, list):
                    ans = ans[key]
                else:
                    raise Exception(f'Unknown type "{type(ans).__name__}"')
            ans[keys[-1]] = value
        elif isinstance(keys, list):
            # TODO: DeprecationWarning
            self.__setitem__(tuple(keys), value)
        else:
            self._data[keys] = value

    def __bool__(self) -> bool:
        return bool(self._data)  # 'list' object has no attribute '__bool__'

    def __iter__(self) -> t.Iterator[t.Any]:
        return self._data.__iter__()

    def __len__(self) -> int:
        return self._data.__len__()

    def __repr__(self) -> str:
        return f'Data({self._data!r})'

    def __str__(self) -> str:
        return self._data.__str__()

    @classmethod
    def from_any(cls, data: FoamItem) -> 'Self':
        return cls(data)

    @classmethod
    def from_dict(cls, data: t.Optional[Dict] = None) -> 'Self':
        return cls({} if data is None else data)

    @classmethod
    def from_dict_keys(cls, *keys: t.Hashable, default: t.Callable = dict) -> 'Self':
        self = cls.from_dict()
        for key in keys:
            self[key] = default()
        return self

    @classmethod
    def from_list(cls, data: t.Optional[List] = None) -> 'Self':
        return cls([] if data is None else data)

    @classmethod
    def from_list_length(cls, length: int, default: t.Callable = lambda: None) -> 'Self':
        return cls.from_list([default() for _ in range(length)])

    @classmethod
    def load(cls, *paths: Path) -> t.Iterator['Self']:
        for path in map(p.Path, paths):
            yield cls.loads(path.read_bytes(), path.suffix[1:])

    @classmethod
    def load_from_path(cls, *parts: str) -> 'Self':
        return next(cls.load(p.Path(*parts)))

    @classmethod
    def loads(cls, content: bytes, type: str = 'yaml') -> 'Self':
        return cls.from_any(Conversion.from_bytes(content, type, all=True).to_document())

    @property
    def data(self) -> FoamItem:
        return self._data

    def contains(self, *keys: t.Any) -> bool:
        try:
            return self.__contains__(keys)
        except Exception:
            return False

    def get(self, key: t.Any, default: t.Optional[t.Any] = None) -> t.Any:
        if isinstance(self._data, dict):
            return self._data.get(key, default)
        elif isinstance(self._data, list):
            try:
                return self._data[key]
            except (IndexError, TypeError):
                return default
        else:
            raise

    def items(self, with_list: bool = False) -> t.Iterator[t.Tuple[Keys[t.Any], t.Any]]:
        yield from self._items(self._data, with_list=with_list)

    def dump(self, *paths: Path) -> 'Self':
        for path in map(p.Path, paths):
            path.write_bytes(self.dumps(path.suffix[1:]))
        return self

    def dump_to_path(self, *parts: str) -> 'Self':
        return self.dump(p.Path(*parts))

    def dumps(self, type: str = 'yaml', **kwargs: t.Any) -> bytes:
        return Conversion \
            .from_document(self._data) \
            .to_bytes(type, all=True, **kwargs)

    def set_default(self, *keys: t.Any, default: t.Any = None) -> 'Self':
        if keys not in self:
            self[keys] = default
        return self[keys]

    def set_via_dict(self, data: Dict) -> 'Self':
        for keys, value in self.__class__.from_dict(data).items():
            self[keys] = value
        return self

    def _items(
        self,
        data: t.Any, with_list: bool = False, keys: Keys[t.Any] = (),
    ) -> t.Iterator[t.Tuple[Keys[t.Any], t.Any]]:
        if isinstance(data, dict):
            for key, value in data.items():
                yield from self._items(value, with_list, keys+(key, ))
        elif with_list and isinstance(data, list):
            for key, value in enumerate(data):
                yield from self._items(value, with_list, keys+(key, ))
        else:
            yield keys, data




__all__ = ['PostProcess', 'VTK']


import typing as t
import warnings as w

from ...base.lib import numpy, vtkmodules
from ...base.type import Array, Dict, Location, Path

if t.TYPE_CHECKING:
    import vtkmodules as _vtkmodules

    from typing_extensions import Self

    from ...base.core import Foam


class PostProcess:
    '''OpenFOAM post-processing'''

    def __init__(self, foam: 'Foam') -> None:
        self._foam = foam
        self._vtks: t.Optional[t.List['VTK']] = None
        self._logs: t.Optional[Dict] = None

    @classmethod
    def from_foam(cls, foam: 'Foam') -> 'Self':
        return cls(foam)

    @property
    def vtks(self) -> t.List['VTK']:
        if self._vtks is None:
            self._vtks = list(VTK.from_foam(self._foam))
        return self._vtks

    @property
    def logs(self) -> Dict:
        '''Script extract data for each time-step from a log file for graphing

        - Reference:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/bin/foamLog
        '''
        if self._logs is None:
            filename = next(log for log in self._foam.cmd.logs if self._foam.application in log.name).name
            self._foam.cmd.run([f'foamLog {filename}'], overwrite=True, exception=False, unsafe=True)
            self._logs = {
                path.name.replace('_0', ''): numpy.loadtxt(path)
                for path in (self._foam.destination/'logs').iterdir()
                if path.suffix != '.awk'
            }
        return self._logs

    def vtks_set(self, **kwargs: t.Any) -> t.List['VTK']:
        self._vtks = list(VTK.from_foam(self._foam, **kwargs))
        return self._vtks

    def centroid(self, key: str, structured: bool = False) -> t.Dict[float, Array[1, 2]]:
        ans = {}
        for time, vtk in zip(self._foam.cmd.times, self.vtks):
            ans[time] = vtk.centroid(key, structured)
        return ans

    def centroids(
        self,
        keys: t.Optional[t.Set[str]] = None, structured: bool = False,
    ) -> t.Dict[str, t.Dict[float, Array[1, 2]]]:
        return {
            key: self.centroid(key, structured)
            for key in (keys or self._foam.fields)
        }

    def probe(
        self,
        location: Location, keys: t.Optional[t.Set[str]] = None,
        point: bool = True, func: t.Optional[t.Callable] = None,
    ) -> t.Dict[str, t.Dict[float, Array[0, 1]]]:
        location_ = tuple(map(float, location))
        return self.probes(location_, keys=keys, point=point, func=func)[location_]

    def probes(
        self,
        *locations: Location,
        keys: t.Optional[t.Set[str]] = None, point: bool = True, func: t.Optional[t.Callable] = None,
    ) -> t.Dict[Location, t.Dict[str, t.Dict[float, Array[0, 1]]]]:
        ans = {}
        for time, vtk in zip(self._foam.cmd.times, self.vtks):
            probes = vtk.probes(*locations, keys=keys, point=point, func=func)
            for location, probe in probes.items():
                for key, value in probe.items():
                    ans \
                        .setdefault(location, {}) \
                        .setdefault(key, {}) \
                        .setdefault(time, value)
        return ans


class VTK:
    '''OpenFOAM VTK post-processing'''

    def __init__(
        self,
        reader: '_vtkmodules.vtkIOLegacy.vtkDataReader',
        foam: t.Optional['Foam'] = None, point: bool = True, cell: bool = True,
    ) -> None:
        self._foam = foam
        self._points, self._cells = None, None
        self._point_fields = {}
        self._cell_fields = {}
        if point:
            self._points = self._to_numpy(reader.GetOutput().GetPoints().GetData())
            arrays = reader.GetOutput().GetPointData()
            for ith in range(arrays.GetNumberOfArrays()):
                array = arrays.GetArray(ith)
                self._point_fields[array.GetName()] = self._to_numpy(array)
        if cell:
            arrays = reader.GetOutput().GetCellData()
            for ith in range(arrays.GetNumberOfArrays()):
                array = arrays.GetArray(ith)
                self._cell_fields[array.GetName()] = self._to_numpy(array)
            cell_ids = set(self._cell_fields['cellID'])  # TODO: explore why cell_id is sometimes repeated
            for key, value in self._cell_fields.items():
                self._cell_fields[key] = value[:len(cell_ids)]
            self._cells = self._cell_fields['C']

    def __contains__(self, key: str) -> bool:
        return key in self._point_fields or key in self._cell_fields

    def __getitem__(self, key: str) -> None:
        raise NotImplementedError

    @classmethod
    def from_file(cls, path: Path, **kwargs: t.Any) -> 'Self':
        reader = vtkmodules.vtkGenericDataObjectReader()
        reader.SetFileName(str(path))
        for attr in dir(reader):
            if attr.startswith('ReadAll') and attr.endswith('On'):
                getattr(reader, attr)()
        reader.Update()
        self = cls(reader, **kwargs)
        reader.CloseVTKFile()
        return self

    @classmethod
    def from_foam(
        cls,
        foam: 'Foam', options: str = '', overwrite: bool = False,
        **kwargs: t.Any,
    ) -> t.Iterator['Self']:
        foam.destination  # assert dest is not None
        for name in ['writeCellCentres', 'writeCellVolumes']:
            foam.cmd.run([f'postProcess -func {name}'], suffix=f'.{name}', overwrite=overwrite, exception=False, unsafe=True)
        foam.cmd.run([f'foamToVTK {options}'], overwrite=overwrite, exception=False, unsafe=True)
        paths = [
            path
            for path in (foam.destination/'VTK').iterdir()
            if path.is_file() and path.suffix=='.vtk'
        ]
        for path in sorted(paths, key=lambda p: int(p.stem.rsplit('_', maxsplit=1)[-1])):
            yield cls.from_file(path, foam=foam, **kwargs)

    @property
    def foam(self) -> 'Foam':
        assert self._foam is not None

        return self._foam

    @property
    def points(self) -> Array[2]:
        assert self._points is not None

        return self._points

    @property
    def cells(self) -> Array[2]:
        assert self._cells is not None

        return self._cells

    @property
    def point_fields(self) -> t.Dict[str, Array[1, 2]]:
        assert self._point_fields is not None

        return self._point_fields

    @property
    def cell_fields(self) -> t.Dict[str, Array[1, 2]]:
        assert self._cell_fields is not None

        return self._cell_fields

    @property
    def x(self) -> Array[1]:
        return self.points[:, 0]

    @property
    def y(self) -> Array[1]:
        return self.points[:, 1]

    @property
    def z(self) -> Array[1]:
        return self.points[:, 2]

    def keys(self) -> None:
        raise NotImplementedError

    def centroid(self, key: str, structured: bool = False) -> Array[1, 2]:
        if structured:
            coords = self._points
            field = self._point_fields[key]
        else:
            coords = self._cells
            if self._cell_fields[key].ndim == 1:
                weights = self._cell_fields['V']
            else:  # ndim == 2
                weights = self._cell_fields['V'][:, None]
            field = self._cell_fields[key] * weights
        if field.ndim != 1:
            w.warn(f'Array2D is not yet guaranteed: {key}')
        return (coords.T @ field) / sum(field)

    def centroids(
        self,
        keys: t.Optional[t.Set[str]] = None, structured: bool = False,
    ) -> t.Dict[str, Array[1, 2]]:
        return {
            key: self.centroid(key, structured)
            for key in (keys or self.foam.fields)
        }

    def centroid_with_args(self, *keys: str, structured: bool = False) -> t.Dict[str, Array[1, 2]]:
        return self.centroids(set(keys), structured=structured)

    def probe(
        self,
        location: Location, keys: t.Optional[t.Set[str]] = None,
        point: bool = True, func: t.Optional[t.Callable] = None,
    ) -> t.Dict[str, Array[0, 1]]:
        location_ = tuple(map(float, location))
        return self.probes(location_, keys=keys, point=point, func=func)[location_]

    def probes(
        self,
        *locations: Location,
        keys: t.Optional[t.Set[str]] = None, point: bool = True, func: t.Optional[t.Callable] = None,
    ) -> t.Dict[Location, t.Dict[str, Array[0, 1]]]:
        '''
        - Reference:
            - https://github.com/OpenFOAM/OpenFOAM-7/tree/master/src/sampling/probes
        '''
        keys = keys or self.foam.fields
        coords = self.points if point else self.cells
        fields = self.point_fields if point else self.cell_fields
        func = func or (lambda x: numpy.square(x).mean(axis=1))
        ans = {}
        for location in locations:
            index = numpy.argmin(func(coords-location))
            ans[tuple(map(float, location))] = {
                key: fields[key][index]
                for key in keys
            }

    def _to_numpy(self, array: '_vtkmodules.vtkCommonCore.vtkDataArray') -> Array[1, 2]:
        return vtkmodules.vtk_to_numpy(array)



__all__ = ['lib']


import types
import typing as t
import warnings as w

from ..base.type import CachedLib


class lib:
    '''Load modules on demand'''

    _cache: t.Dict[str, CachedLib] = {}

    def __class_getitem__(cls, key: str) -> t.Optional[CachedLib]:
        if key in cls._cache:
            ans = cls._cache[key]
        else:
            attr = getattr(cls, key)
            try:
                ans = attr()
            except (ModuleNotFoundError, ImportError) as e:
                ans = None
                w.warn(f'{e.msg} ({attr.__doc__})')
            cls._cache[key] = ans
        return ans

    @classmethod
    def reset(cls) -> type:
        cls._cache.clear()
        return cls

    @classmethod
    def click(cls) -> types.ModuleType:
        '''https://pypi.org/project/click'''
        import click

        return click

    @classmethod
    def lark(cls) -> types.ModuleType:
        '''https://pypi.org/project/lark'''
        import lark

        return lark

    @classmethod
    def numpy(cls) -> types.ModuleType:
        '''https://pypi.org/project/numpy'''
        import numpy

        return numpy

    @classmethod
    def py7zr(cls) -> types.ModuleType:
        '''https://pypi.org/project/py7zr'''
        import py7zr

        return py7zr

    @classmethod
    def tqdm(cls) -> types.ModuleType:
        '''https://pypi.org/project/tqdm'''
        import tqdm

        return tqdm

    @classmethod
    def vtk(cls) -> types.ModuleType:
        '''https://pypi.org/project/vtk'''
        import vtkmodules.all as vtk

        return vtk

    @classmethod
    def yaml(cls) -> types.ModuleType:
        '''https://pypi.org/project/PyYAML'''
        import yaml

        return yaml

    @classmethod
    def SafeLoader(cls) -> object:
        '''https://pypi.org/project/PyYAML'''
        try:
            from yaml import CSafeLoader as SafeLoader
        except ImportError:
            from yaml import SafeLoader

        return SafeLoader

    @classmethod
    def vtk_to_numpy(cls) -> t.Callable:
        '''https://pypi.org/project/vtk'''
        from vtkmodules.util.numpy_support import vtk_to_numpy

        return vtk_to_numpy
